#![no_std]
#![no_main]

mod crypto;
mod datetime;

use arduino_hal::default_serial;
use arduino_hal::delay_ms;
use crypto::hmac_sha1;

use panic_halt as _;

const INTERVAL: u64 = 30;
const SECRET_KEY_MAX_LEN: usize = 32;

fn process_secret<'a>(secret: &str, output: &mut [u8]) -> usize {
    // RFC 6238 Doesn't specify a max length for the secret keys, So I chose 32
    // You can change the max length by modifying the const
    let mut result = [0u8; SECRET_KEY_MAX_LEN];
    let mut inc = 0;
    let offset = match 8 - (secret.len() % 8) {
        8 => 0,
        n => n,
    };

    for c in secret.chars() {
        result[inc] = c.to_ascii_uppercase() as u8;
        inc += 1;
    }
    for _ in 0..offset {
        result[inc] = '=' as u8;
        inc += 1;
    }

    let decoded = match binascii::b32decode(&result[..inc], output).ok() {
        Some(bytes) => bytes,
        None => "".as_bytes(), // TODO: Implement a way to display errors
    };
    decoded.len()
}

fn generate_otp_token(private_key: &str, actual_time: u64) -> u32 {
    let mut output = [0u8; 32];
    let private_key_len = process_secret(private_key, &mut output);

    let interval = actual_time / INTERVAL;
    let msg = interval.to_be_bytes();

    let mut hmac_digest = [0u8; 20];
    hmac_sha1(&output[..private_key_len], &msg, &mut hmac_digest);

    let start = (hmac_digest[19] & 0xF) as usize;
    let bytes: [u8; 4] = hmac_digest[start..start + 4].try_into().unwrap();

    let raw_token = u32::from_be_bytes(bytes);
    ((raw_token & 0x7F_FF_FF_FF) % 1_000_000) as u32
}

#[arduino_hal::entry]
fn main() -> ! {
    let dp = arduino_hal::Peripherals::take().unwrap();
    let pins = arduino_hal::pins!(dp);

    let mut led = pins.d13.into_output();
    let mut serial = default_serial!(dp, pins, 9600);

    let i2c_clock = arduino_hal::I2c::new(
        dp.TWI,
        pins.a4.into_pull_up_input(),
        pins.a5.into_pull_up_input(),
        50000,
    );
    let mut rtc = ds323x::Ds323x::new_ds3231(i2c_clock);

    let datetime = datetime::Datetime::from_ds3231(&mut rtc);

    let timestamp = datetime.unix_epoch();
    ufmt::uwriteln!(&mut serial, "Timestamp: {}", timestamp).unwrap();

    let token = generate_otp_token("FS7J22EHLLSOGKUVJKV2XE7FTIX24JAJ", 1680155315);
    ufmt::uwriteln!(&mut serial, "token: {}", token).unwrap();

    delay_ms(5000);
    let token = generate_otp_token("FS7J22EHLLSOGKUVJKV2XE7FTIX24JAJ", 1680155315);
    ufmt::uwriteln!(&mut serial, "token: {}", token).unwrap();

    delay_ms(5000);
    let token = generate_otp_token("FS7J22EHLLSOGKUVJKV2XE7FTIX24JAJ", 1680155315);
    ufmt::uwriteln!(&mut serial, "token: {}", token).unwrap();

    loop {
        led.toggle();
        delay_ms(1000);
    }
}
